(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_netfs.ml 387 2011-09-05 13:36:59Z gerd $ *)

open Printf
open Uq_engines.Operators

let tr_errno e =
  match e with
    | `enotrans      -> Unix.EPERM
    | `efailedcommit -> Unix.EIO
    | `elongtrans    -> Unix.EIO
    | `efailed       -> Unix.EIO
    | `eperm         -> Unix.EPERM
    | `enoent        -> Unix.ENOENT
    | `eaccess       -> Unix.EACCES
    | `eexist        -> Unix.EEXIST
    | `efhier        -> Unix.EINVAL
    | `einval        -> Unix.EINVAL
    | `efbig         -> Unix.EFBIG
    | `enospc        -> Unix.ENOSPC
    | `erofs         -> Unix.EROFS
    | `enametoolong  -> Unix.ENAMETOOLONG
    | `econflict     -> Unix.EPERM
    | `ecoord        -> Unix.EPERM
    | `enonode       -> Unix.EIO
    | `etbusy        -> Unix.EPERM
    | `eio           -> Unix.EIO
    | `estale        -> (* Unix.ESTALE *) Unix.ENOENT
    | `eloop         -> Unix.ELOOP
    | `enotdir       -> Unix.ENOTDIR
    | `eisdir        -> Unix.EISDIR
    | `enotempty     -> Unix.ENOTEMPTY
    | `ebadpath      -> Unix.ENOTDIR

let tr_error e fn file =
  match e with
    | Plasma_client.Plasma_error errno ->
	Unix.Unix_error(tr_errno errno, fn, file)
    | _ ->
	e


let catch_error fn file f arg =
  try
    f arg
  with
    | e ->
	raise(tr_error e fn file)


let w_trans fn file cluster f =
  (* Includes support for retry and for catch_error *)
  catch_error fn file
    (Plasma_client.retry cluster file
       (fun () ->
	  Plasma_client.with_trans cluster f
       )
    )
    ()


let check_is_regular cluster inode =
  let ii = Plasma_client.get_cached_inodeinfo cluster inode false in
          (* also refreshes metadata *)
  match ii.Plasma_rpcapi_aux.filetype with
    | `ftype_regular -> ()
    | `ftype_directory ->
	raise(Plasma_client.Plasma_error `eisdir)
    | _ -> assert false


let input_file_1 cluster inode file init_pos =
  let () = check_is_regular cluster inode in
  let pos = ref init_pos in
  let closed = ref false in
object(self)
  method input s p len =
    if !closed then raise Netchannels.Closed_channel;
    if len = 0 then
      0
    else (
      let (n,eof,_ii) =
	catch_error "Plasma_netfs.read" file
	  (Plasma_client.read cluster inode !pos (`String s) p) len in
      if n=0 && eof then raise End_of_file;
      pos := Int64.add !pos (Int64.of_int n);
      n
    )
  method close_in() =
    closed := true
end


let input_file cluster inode file init_pos =
  Netchannels.lift_in (`Rec (input_file_1 cluster inode file init_pos))


let output_file_1 cluster inode file =
  let () = check_is_regular cluster inode in
  let pos = ref 0L in
  let closed = ref false in
object(self)
  method output s p l =
    if !closed then raise Netchannels.Closed_channel;
    if l = 0 then
      0
    else (
      let n =
	catch_error "Plasma_netfs.write" file
	  (Plasma_client.write cluster inode !pos (`String s) p) l in
      pos := Int64.add !pos (Int64.of_int n);
      n
    )

  method flush() =
    if !closed then raise Netchannels.Closed_channel;
    catch_error "Plasma_netfs.write" file
      (Plasma_client.flush cluster inode 0L) 0L

  method close_out() =
    if not !closed then (
      self # flush();
      closed := true
    )
end


let output_file cluster inode file =
  Netchannels.lift_out
    ~buffered:false
    (`Rec (output_file_1 cluster inode file))

let list_isect_empty l1 l2 = (* whether intersection is empty *)
  List.for_all
    (fun x1 -> not (List.mem x1 l2))
    l1

let slash_re = Netstring_pcre.regexp "/+"


class netfs ?(repl=0) ?(verbose=false) cluster : Netfs.stream_fs =
  let fix_ii ii =
    { ii with 
	Plasma_rpcapi_aux.replication = repl } in
object(self)
  inherit Netfs.empty_fs "Plasma_netfs"

  method path_encoding = Some `Enc_utf8
  method path_exclusions = [ 0,0; 47,47 ]
  method nominal_dot_dot = false

  method read flags name =
    catch_error "Plasma_netfs.read" name
      (fun () ->
	 let inode =
	   Plasma_client.with_trans
	     cluster
	     (fun trans -> Plasma_client.lookup trans name false) in
	 let skip_d = 
	   try
             List.find
               (fun flag -> 
		  match flag with
		    | `Skip _ -> true
		    | _ -> false
               ) 
               flags 
	   with Not_found -> `Skip 0L in
	 let skip =
	   match skip_d with
             | `Skip n -> n
             | _ -> assert false in
	 input_file cluster inode name skip
      )
      ()

  method write flags name =
    let trunc_flag = List.mem `Truncate flags in
    let create_flag = List.mem `Create flags in
    let excl_flag = List.mem `Exclusive flags in

    let inode =
      w_trans "Plasma_netfs.write" name cluster
	(fun trans -> 
	   try
	     let inode = Plasma_client.lookup trans name false in
	     if create_flag && excl_flag then
	       raise(Plasma_client.Plasma_error `eexist);
	     if trunc_flag then (
	       Plasma_client.truncate trans inode 0L;
	       if verbose then
		 eprintf "Truncated: %s\n%!" name;
	       inode
	     )
	     else
	       inode
	   with
	     | Plasma_client.Plasma_error `enoent ->
		 let inode =
		   Plasma_client.create_file trans name 
		     (fix_ii
			(Plasma_client.regular_ii cluster 0o666)) in
		 if verbose then
		   eprintf "Created: %s\n%!" name;
		 inode
	) in
    output_file cluster inode name

  method size _ name =
    w_trans "Plasma_netfs.size" name cluster
      (fun trans -> 
	 let inode = Plasma_client.lookup trans name false in
	 let ii = Plasma_client.get_inodeinfo trans inode in
	 ii.Plasma_rpcapi_aux.eof
      )

  method private test_list_NH trans flags name =
    try
      let inode = Plasma_client.lookup trans name true in
      let ii = Plasma_client.get_inodeinfo trans inode in
      if ii.Plasma_rpcapi_aux.filetype = `ftype_symlink then
        [ `N; `H ]
      else
        [ `N ]
    with
      | Plasma_client.Plasma_error(`enoent | `ebadpath) -> []
	  
  method private test_list_EDFS trans flags name =
    try
      let inode = Plasma_client.lookup trans name (List.mem `Link flags) in
      let ii = Plasma_client.get_inodeinfo trans inode in
      let non_empty = ii.Plasma_rpcapi_aux.eof <> 0L in
      let kind_l =
        match ii.Plasma_rpcapi_aux.filetype with
          | `ftype_regular -> [ `F ]
          | `ftype_directory -> [ `D ]
          | _ -> [] in
      [ `E ] @ kind_l @ (if non_empty then [`S] else [])
    with
      | Plasma_client.Plasma_error(`enoent | `ebadpath) -> []
	  
  method test flags name ttype =
    w_trans "Plasma_netfs.test" name cluster
      (fun trans -> 
	 let l =
	   match ttype with
             | `N | `H -> self#test_list_NH trans flags name
             | `E | `D | `F | `S -> self#test_list_EDFS trans flags name
             | `R | `W | `X -> [] in
         List.mem ttype l
      )

  method test_list flags name tests =
    w_trans "Plasma_netfs.test_list" name cluster
      (fun trans -> 
         let nh =
           if not(list_isect_empty tests [`N;`H]) then
             self#test_list_NH trans flags name
           else
             [] in
         let edfs =
           if not(list_isect_empty tests [`E;`D;`F;`S]) then
             self#test_list_EDFS trans flags name
           else
             [] in
         let rwx = [] in
         List.map
           (fun t ->
              match t with
		| `N | `H -> List.mem t nh
		| `E | `D | `F | `S -> List.mem t edfs
		| `R | `W | `X -> List.mem t rwx
           )
           tests
      )

  method remove flags name = (* `Recursive *)
    let recurse = List.mem `Recursive flags in
    w_trans "Plasma_netfs.remove" name cluster
      (fun trans -> 
	 let rec remove_dir_contents dir_name dir_inode =
	   let entries = Plasma_client.list_inode trans dir_inode in
	   List.iter
	     (fun (base,_) -> 
		remove_entry (dir_name ^ "/" ^ base) dir_inode base)
	     entries
    
	 and remove_entry name dir_inode base =
	   Plasma_client.namelock trans dir_inode base;
	   try
	     Plasma_client.unlink_at trans dir_inode base;
	     if verbose then
	       eprintf "Removed: %s\n%!" name
	   with
	     | Plasma_client.Plasma_error `enotempty when recurse ->
		 let inode = 
		   Plasma_client.dir_lookup trans dir_inode base true in
		 let ii = Plasma_client.get_inodeinfo trans inode in
		 if ii.Plasma_rpcapi_aux.filetype = `ftype_directory then
		   remove_dir_contents name inode;
		 Plasma_client.unlink_at trans dir_inode base;
		 if verbose then
		   eprintf "Removed: %s\n%!" name
	 in
	 
	 let dir = Filename.dirname name in
	 let base = Filename.basename name in
	 let dir_inode = Plasma_client.lookup trans dir false in
	 remove_entry name dir_inode base
      );
    if verbose then eprintf "Committed.\n%!"
    

  method rename flags oldname newname =
    w_trans "Plasma_netfs.rename" oldname cluster
      (fun trans -> 
	 Plasma_client.rename trans oldname newname;
	 if verbose then
	   eprintf "Renamed: %s -> %s\n%!" oldname newname
      );
    if verbose then eprintf "Committed.\n%!"


  method symlink flags oldname newname =
    w_trans "Plasma_netfs.symlink" newname cluster
      (fun trans -> 
	 ignore(
	   Plasma_client.create_file trans newname 
	     (Plasma_client.symlink_ii cluster oldname));
	 if verbose then
	   eprintf "Linked: %s -> %s\n%!" newname oldname
      );
    if verbose then eprintf "Committed.\n%!"


  method readdir flags name =
    w_trans "Plasma_netfs.readdir" name cluster
      (fun trans -> 
	 [ "."; ".." ] @
	   (List.map 
	      fst
	      (Plasma_client.list trans name)
	   )
      )

  method readlink flags name =
    w_trans "Plasma_netfs.readlink" name cluster
      (fun trans -> 
	 let inode = Plasma_client.lookup trans name true in
	 let ii = Plasma_client.get_inodeinfo trans inode in
	 match ii.Plasma_rpcapi_aux.filetype with
	   | `ftype_symlink ->
	       ii.Plasma_rpcapi_aux.field1
	   | _ ->
	       raise(Plasma_client.Plasma_error `einval)
      )    

  method mkdir flags name = (* `Path, `Nonexcl *)
    w_trans "Plasma_netfs.mkdir" name cluster
      (fun trans ->
	 if List.mem `Path flags then
	   self#mkdir_p trans name
	 else (
	   try
	     ignore(
	       Plasma_client.mkdir trans name 
		 (Plasma_client.dir_ii cluster 0o777));
	     if verbose then
	       eprintf "Created dir: %s\n%!" name
	   with
             | Plasma_client.Plasma_error `eexist 
		 when List.mem `Nonexcl flags ->
		 ()
	 )
      );
    if verbose then eprintf "Committed.\n%!"


  method private mkdir_p trans filename =
    let rec traverse curdir todo =
      match todo with
        | [] -> ()
        | d :: todo' ->
            let curdir' = curdir @ [d] in
            let p = String.concat "/" curdir' in
            ( try 
		ignore(
		  Plasma_client.mkdir trans p
		    (Plasma_client.dir_ii cluster 0o777));
		if verbose then
		  eprintf "Created dir: %s\n%!" p
	      with 
		| Plasma_client.Plasma_error `eexist -> ()
            );
            traverse curdir' todo' in
    let l = Netstring_pcre.split_delim slash_re filename in
    match l with
      | "" :: tl ->
          traverse [""] tl
      | _ ->
	  raise(Plasma_client.Plasma_error `einval)
	    
  method rmdir flags name =
    w_trans "Plasma_netfs.rmdir" name cluster
      (fun trans -> 
	 let inode = Plasma_client.lookup trans name true in
	 let ii = Plasma_client.get_inodeinfo trans inode in
	 match ii.Plasma_rpcapi_aux.filetype with
	   | `ftype_directory ->
	       Plasma_client.unlink trans name;  (* FIXME (potential race) *)
	       if verbose then
		 eprintf "Removed dir: %s\n%!" name
	   | _ ->
	       raise(Plasma_client.Plasma_error `enotdir)
      );
    if verbose then eprintf "Committed.\n%!"


  method copy flags oldname newname =
    (* FIXME: This is a dumb implementation. We could avoid the pipe.
       Ideally, the datanodes storing the original version of the file
       would directly send the blocks to the datanodes for the copy
     *)
    let esys = Plasma_client.event_system cluster in
    let fd_rd, fd_wr = Unix.pipe() in
    Unix.set_nonblock fd_rd;
    Unix.set_nonblock fd_wr;
    let fd_wr_closed = ref false in

    let copy_e oldinode newinode len =
      let e1 =
	Plasma_client.copy_out_e cluster oldinode 0L fd_wr len 
	++ (fun _ ->
	      Unix.close fd_wr;
	      fd_wr_closed := true;
	      eps_e (`Done ()) esys 
	   ) in
      let e2 = 
	Plasma_client.copy_in_e 
	  ~flags:[`Late_datasync]
	  cluster newinode 0L fd_rd len `Star in
      Uq_engines.sync_engine e1 e2 in

    let e =
      Plasma_client.retry_e cluster "Plasma_netfs.copy"
	(fun () ->
	   Plasma_client.with_trans_e cluster
	     (fun trans ->
		Plasma_client.lookup_e trans oldname false
		++ (fun oldinode ->
		      Plasma_client.get_inodeinfo_e trans oldinode
		      ++ (fun oldii ->
			    let oldft = oldii.Plasma_rpcapi_aux.filetype in
			    if oldft <> `ftype_regular then
			      raise(Plasma_client.Plasma_error `eisdir);
			    Uq_engines.meta_engine
			      (Plasma_client.lookup_e trans newname false)
			    ++ (function
				  | `Error(Plasma_client.Plasma_error _) ->
				      Plasma_client.create_file_e trans newname 
					(fix_ii
					   (Plasma_client.regular_ii 
					      cluster 0o666))
				  | `Done newinode ->
				      Plasma_client.truncate_e trans newinode 0L
				      ++ (fun () ->
					    eps_e (`Done newinode) esys
					 )
				  | st -> 
				      eps_e
					(st :> _ Uq_engines.engine_state) esys
			       )
			    ++ (fun newinode ->
				  let len = oldii.Plasma_rpcapi_aux.eof in
				  eps_e (`Done (oldinode, newinode, len)) esys
			       )
			 )
		   )
	     )
	)
	()
      ++ (fun (oldinode, newinode, len) ->
	    copy_e oldinode newinode len
	 )
    in
    Unixqueue.run esys;
    if not !fd_wr_closed then Unix.close fd_wr;
    Unix.close fd_rd;
    match e # state with
      | `Done _ -> 
	  if verbose then
	    eprintf "Copied: %s -> %s\n%!" oldname newname
      | `Error e -> raise (tr_error e "Plasma_netfs.copy" oldname)
      | `Aborted -> failwith "aborted"
      | `Working _ -> assert false
end


let netfs = new netfs
